==> ec2init-0.17-py3-none-any.whl: ec2init/plans/preinit/format_nvme_drives.py <==
import json
import logging
import pathlib
import time
import re

from typing import List, Optional

from ec2init import Procedure
from ec2init.plans.init.start_services import StartServices
from ec2init.utils.system import OSType, SystemUtils

logger = logging.getLogger(__name__)

class FormatNVMeDrives(Procedure):
    def __init__(self, name: str, check_file: Optional[pathlib.Path]) -> None:
        self.name = name
        self.check_file = check_file

    def execute(self) -> List[Procedure]:
        logger.info(f"{self.name} - Starting execution")
        nvme_devices = self._get_nvme_devices()
        if not nvme_devices:
            logger.info(f"{self.name} - Did not find any (unpartitioned) NVMe drive, skipping")
            self._done()
            return []

        partitions = []
        for dev in nvme_devices:
            partition = self._partition_device(dev)
            if partition:
                partitions.append(partition)

        for partition in partitions:
            self._format_xfs_partition(partition)

        self._done()
        logger.info(f"{self.name} - Finished execution")
        return []

    def rollback(self) -> None:
        pass
        
    def _get_nvme_devices(self, partitioned: bool = False) -> List[str]:
        os_type = SystemUtils.get_os_type()
        if os_type == OSType.UBUNTU:
            return self._get_ubuntu_nvme_drives(partitioned)
        if os_type == OSType.CENTOS:
            return self._get_centos_nvme_drives(partitioned)
        raise Exception(f"Unsupported OS type {os_type}")

    def _get_ubuntu_nvme_drives(self, partitioned: bool) -> List[str]:
        result = SystemUtils.execute_command(['lsblk', '-J', '-o', '+SERIAL'])
        dev_js = json.loads(result.stdout)
        devices = dev_js['blockdevices']
        nvme_devs = []
        for dev in devices:
            # Serial number of NVMe drives in AWS start with AWS
            # while EBS drives have the volume ID volxxxxxxxx
            if dev['serial'] and dev['serial'].startswith('AWS'):
                if partitioned:
                    nvme_devs.append(dev['name'])
                # If a drive is partitioned it will have a child
                elif not partitioned and 'children' not in dev:
                    nvme_devs.append(dev['name'])

        return [f"/dev/{d}" for d in nvme_devs]

    def _get_centos_nvme_drives(self, partitioned: bool) -> List[str]:
        result = SystemUtils.execute_command(['lsblk', '-nlo', '+SERIAL'])
        stdout = result.stdout.decode('utf-8').strip('\n')
        devs = stdout.split('\n')
        nvme_devs = []
        idx = 0 
        for dev in devs:
            if 'AWS' in dev:
                # Only if 'AWS' is included in the Serial number
                # Replace spaces with '-', remove last '-' and then split it
                splitted_dev = re.sub(r'\s+', '-', dev).strip('-').split('-')
                if partitioned:
                    # If we want to include partitioned devices as well
                    # include the device name indiscriminately
                    nvme_devs.append(splitted_dev[0])
                else:
                    # In lsblk output first comes the device
                    # and then the partition IF the device is partioned
                    # For example:
                    # NAME           MAJ:MIN RM   SIZE RO TYPE MOUNTPOINT SERIAL
                    # /dev/nvme0n1   259:0    0 442.4G  0 disk            AWS7B5BC4CFEC5580FEA
                    # /dev/nvme0n1p1 259:1    0 442.4G  0 part
                    #
                    # We do not want partitioned devices so we should check n+1 line, if exists
                    splitted_dev_next = ['']
                    if idx < len(devs) - 1:
                        splitted_dev_next = re.sub(r'\s+', '-', devs[idx+1]).strip('-').split('-')
                    if splitted_dev[0] not in splitted_dev_next[0]:
                        if len(splitted_dev) == 7:
                            # If we don't want partitioned devices the output
                            # should have 7 splits, partition split should be missing
                            nvme_devs.append(splitted_dev[0])
            idx += 1
        return [f"/dev/{d}" for d in nvme_devs]


    def _partition_device(self, device: str) -> Optional[str]:
        logger.debug(f"{self.name} - Partitioning {device}")
        partition = self._get_device_partition(device)
        if partition:
            logger.info(f"{self.name} - Device {device} is already partitioned, skipping...")
            return None

        logger.info(f"{self.name} - Partitioning device {device}")
        SystemUtils.execute_command([f"echo ';' | sfdisk {device}"], shell=True)
        logger.info(f"{self.name} - Partitioned device {device}")
        # Give it some time
        time.sleep(10)
        partition = self._get_device_partition(device)
        if not partition:
            raise Exception(f"Device {device} expected to have partition but it does not")
        return partition

    def _get_ubuntu_device_partition(self, device: str) -> Optional[str]:
        result = SystemUtils.execute_command(['lsblk', '-J', device])
        res_js = json.loads(result.stdout)
        block_device = res_js['blockdevices'][0]
        if 'children' not in block_device or not block_device['children']:
            logger.info(f"{self.name} - Device {device} does not have partitions")
            return None
        # We assume the device has only one partition
        partition = block_device['children'][0]['name']
        return f"/dev/{partition}"

    def _get_centos_device_partition(self, device: str) -> Optional[str]:
        result = SystemUtils.execute_command(['lsblk', '-nl', device])
        stdout = result.stdout.decode('utf-8').strip('\n')
        devs = stdout.split('\n')
        if len(devs) == 1:
            logger.info(f"{self.name} - Device {device} does not have partitions")
            return None
        # We assume only one partition
        partition_split = re.sub(r'\s+', '-', devs[1]).strip('-').split('-')
        return f"/dev/{partition_split[0]}"

    def _get_device_partition(self, device: str) -> Optional[str]:
        os_type = SystemUtils.get_os_type()
        if os_type == OSType.UBUNTU:
            return self._get_ubuntu_device_partition(device)
        if os_type == OSType.CENTOS:
            return self._get_centos_device_partition(device)
        raise Exception(f"Unsupported OS type {os_type}")

    def _format_xfs_partition(self, partition: str) -> None:
        logger.info(f"{self.name} - Formatting {partition} to XFS")
        SystemUtils.execute_command(['mkfs.xfs', partition])
        logger.info(f"{self.name} - Formatted {partition} to XFS")

    def _done(self) -> None:
        if self.check_file:
            self.check_file.touch()

class WorkerFormatNVMEDrives(FormatNVMeDrives):
    def __init__(self, name: str, check_file: pathlib.Path) -> None:
        super().__init__(name, None)
        self.check_file = check_file

    def execute(self) -> List[Procedure]:
        logger.info(f"{self.name} - Executing")
        super().execute()
        nvme_devices = self._get_nvme_devices(partitioned=True)
        if not nvme_devices:
            logger.info(f"{self.name} - Did not find any NVMe drive, skipping")
            self._done()
            return []
        nvme_dev = nvme_devices[0]
        partition = self._get_device_partition(nvme_dev)
        if not partition:
            err_msg = f"{self.name} - Expected {nvme_dev} to have been partitioned but it is not!"
            logger.error(err_msg)
            raise Exception(err_msg)
        mount_point = self._create_mount_point(partition)
        self._mount_partition(partition, mount_point)
        self._reorganize_directories(mount_point)

        self._done()
        logger.info(f"{self.name} - Finished execution")
        return []

    def _create_mount_point(self, partition: str) -> pathlib.Path:
        name = pathlib.Path(partition).name
        root = pathlib.Path('/mnt')
        mount_point = root / name
        logger.debug(f"{self.name} - Creating mount point {str(mount_point)}")
        mount_point.mkdir()
        logger.info(f"{self.name} - Created mount point {str(mount_point)}")
        return mount_point

    def _mount_partition(self, partition: str, mount_point: pathlib.Path) -> None:
        logger.debug(f"{self.name} - Mounting {partition} to {str(mount_point)}")
        post_exception_action = lambda: SystemUtils.execute_command(["umount", "-f", str(mount_point)], timeout=500)
        SystemUtils.execute_command_with_retry(["mount", partition, str(mount_point)], post_exception_action=post_exception_action)
        logger.info(f"{self.name} - Mounted {partition} to {str(mount_point)}")

    def _reorganize_directories(self, mount_point: pathlib.Path) -> None:
        logger.debug(f"{self.name} - Moving /srv/hops/hopsdata to {mount_point}")
        hopsdata = pathlib.Path('/srv/hops/hopsdata')
        real_hopsdata = hopsdata.resolve()
        cmd = f"mv -f {str(real_hopsdata)} {str(mount_point)}"
        # We have to use mv because Python shutil does not preserve metadata
        # when moving across different filesystems
        SystemUtils.execute_command(cmd.split(" "), timeout=600)
        logger.info(f"{self.name} - Moved /srv/hops/hopsdata to {mount_point}")

        new_hopsdata = mount_point / 'hopsdata'
        real_hopsdata.symlink_to(new_hopsdata)
        logger.info(f"{self.name} - symlinked {str(real_hopsdata)} to {str(new_hopsdata)}")
        logger.info(f"{self.name} - Finished moving hopsdata to NVMe drive")

        logger.debug(f"{self.name} - Moving /srv/hops/docker to {mount_point}")
        docker_dir = pathlib.Path('/srv/hops/docker')
        real_docker_dir = docker_dir.resolve()
        docker_size = sum(f.stat().st_size for f in real_docker_dir.glob('**/*') if f.is_file())
        if docker_size <= 100 * 1024 * 1024:
            is_docker_running = StartServices._is_service_running('docker')
            logger.debug(f"{self.name} - Is docker service running? {is_docker_running}")
            if is_docker_running:
                StartServices.stop_service('docker')

            cmd = f"mv -f {str(real_docker_dir)} {str(mount_point)}"
            SystemUtils.execute_command(cmd.split(" "), timeout=1800)

            new_docker_dir = mount_point / 'docker'
            real_docker_dir.symlink_to(new_docker_dir)
            logger.info(f"{self.name} - symlinked {str(real_docker_dir)} to {str(new_docker_dir)}")
            if is_docker_running:
                logger.debug(f"{self.name} - Starting docker service")
                StartServices.start_service('docker')
            logger.info(f"{self.name} - Finished moving docker to NVMe drive")
        else:
            logger.info(f"{self.name} - Skip moving docker directory to NVMe drive because it is too big")


    def rollback(self) -> None:
        pass

==> ec2init-0.19.1-py3-none-any.whl: ec2init/plans/preinit/format_nvme_drives.py <==
import json
import logging
import pathlib
import time
import re

from typing import List, Optional

from ec2init import Procedure
from ec2init.plans.init.start_services import StartServices
from ec2init.utils.system import OSType, SystemUtils
from ec2init.utils.systemd import SystemdUtils

logger = logging.getLogger(__name__)

class FormatNVMeDrives(Procedure):
    def __init__(self, name: str, check_file: Optional[pathlib.Path]) -> None:
        self.name = name
        self.check_file = check_file

    def execute(self) -> List[Procedure]:
        logger.info(f"{self.name} - Starting execution")
        nvme_devices = self._get_nvme_devices()
        if not nvme_devices:
            logger.info(f"{self.name} - Did not find any (unpartitioned) NVMe drive, skipping")
            self._done()
            return []

        partitions = []
        for dev in nvme_devices:
            partition = self._partition_device(dev)
            if partition:
                partitions.append(partition)

        for partition in partitions:
            self._format_xfs_partition(partition)

        self._done()
        logger.info(f"{self.name} - Finished execution")
        return []

    def rollback(self) -> None:
        pass
        
    def _get_nvme_devices(self, partitioned: bool = False) -> List[str]:
        os_type = SystemUtils.get_os_type()
        if os_type == OSType.UBUNTU:
            return self._get_ubuntu_nvme_drives(partitioned)
        if os_type == OSType.CENTOS:
            return self._get_centos_nvme_drives(partitioned)
        raise Exception(f"Unsupported OS type {os_type}")

    def _get_ubuntu_nvme_drives(self, partitioned: bool) -> List[str]:
        result = SystemUtils.execute_command(['lsblk', '-J', '-o', '+SERIAL'])
        dev_js = json.loads(result.stdout)
        devices = dev_js['blockdevices']
        nvme_devs = []
        for dev in devices:
            # Serial number of NVMe drives in AWS start with AWS
            # while EBS drives have the volume ID volxxxxxxxx
            if dev['serial'] and dev['serial'].startswith('AWS'):
                if partitioned:
                    nvme_devs.append(dev['name'])
                # If a drive is partitioned it will have a child
                elif not partitioned and 'children' not in dev:
                    nvme_devs.append(dev['name'])

        return [f"/dev/{d}" for d in nvme_devs]

    def _get_centos_nvme_drives(self, partitioned: bool) -> List[str]:
        result = SystemUtils.execute_command(['lsblk', '-nlo', '+SERIAL'])
        stdout = result.stdout.decode('utf-8').strip('\n')
        devs = stdout.split('\n')
        nvme_devs = []
        idx = 0 
        for dev in devs:
            if 'AWS' in dev:
                # Only if 'AWS' is included in the Serial number
                # Replace spaces with '-', remove last '-' and then split it
                splitted_dev = re.sub(r'\s+', '-', dev).strip('-').split('-')
                if partitioned:
                    # If we want to include partitioned devices as well
                    # include the device name indiscriminately
                    nvme_devs.append(splitted_dev[0])
                else:
                    # In lsblk output first comes the device
                    # and then the partition IF the device is partioned
                    # For example:
                    # NAME           MAJ:MIN RM   SIZE RO TYPE MOUNTPOINT SERIAL
                    # /dev/nvme0n1   259:0    0 442.4G  0 disk            AWS7B5BC4CFEC5580FEA
                    # /dev/nvme0n1p1 259:1    0 442.4G  0 part
                    #
                    # We do not want partitioned devices so we should check n+1 line, if exists
                    splitted_dev_next = ['']
                    if idx < len(devs) - 1:
                        splitted_dev_next = re.sub(r'\s+', '-', devs[idx+1]).strip('-').split('-')
                    if splitted_dev[0] not in splitted_dev_next[0]:
                        if len(splitted_dev) == 7:
                            # If we don't want partitioned devices the output
                            # should have 7 splits, partition split should be missing
                            nvme_devs.append(splitted_dev[0])
            idx += 1
        return [f"/dev/{d}" for d in nvme_devs]


    def _partition_device(self, device: str) -> Optional[str]:
        logger.debug(f"{self.name} - Partitioning {device}")
        partition = self._get_device_partition(device)
        if partition:
            logger.info(f"{self.name} - Device {device} is already partitioned, skipping...")
            return None

        logger.info(f"{self.name} - Partitioning device {device}")
        SystemUtils.execute_command([f"echo ';' | sfdisk {device}"], shell=True)
        logger.info(f"{self.name} - Partitioned device {device}")
        # Give it some time
        time.sleep(10)
        partition = self._get_device_partition(device)
        if not partition:
            raise Exception(f"Device {device} expected to have partition but it does not")
        return partition

    def _get_ubuntu_device_partition(self, device: str) -> Optional[str]:
        result = SystemUtils.execute_command(['lsblk', '-J', device])
        res_js = json.loads(result.stdout)
        block_device = res_js['blockdevices'][0]
        if 'children' not in block_device or not block_device['children']:
            logger.info(f"{self.name} - Device {device} does not have partitions")
            return None
        # We assume the device has only one partition
        partition = block_device['children'][0]['name']
        return f"/dev/{partition}"

    def _get_centos_device_partition(self, device: str) -> Optional[str]:
        result = SystemUtils.execute_command(['lsblk', '-nl', device])
        stdout = result.stdout.decode('utf-8').strip('\n')
        devs = stdout.split('\n')
        if len(devs) == 1:
            logger.info(f"{self.name} - Device {device} does not have partitions")
            return None
        # We assume only one partition
        partition_split = re.sub(r'\s+', '-', devs[1]).strip('-').split('-')
        return f"/dev/{partition_split[0]}"

    def _get_device_partition(self, device: str) -> Optional[str]:
        os_type = SystemUtils.get_os_type()
        if os_type == OSType.UBUNTU:
            return self._get_ubuntu_device_partition(device)
        if os_type == OSType.CENTOS:
            return self._get_centos_device_partition(device)
        raise Exception(f"Unsupported OS type {os_type}")

    def _format_xfs_partition(self, partition: str) -> None:
        logger.info(f"{self.name} - Formatting {partition} to XFS")
        SystemUtils.execute_command(['mkfs.xfs', partition])
        logger.info(f"{self.name} - Formatted {partition} to XFS")

    def _done(self) -> None:
        if self.check_file:
            self.check_file.touch()

class WorkerFormatNVMEDrives(FormatNVMeDrives):
    def __init__(self, name: str, check_file: pathlib.Path) -> None:
        super().__init__(name, None)
        self.check_file = check_file

    def execute(self) -> List[Procedure]:
        logger.info(f"{self.name} - Executing")
        super().execute()
        nvme_devices = self._get_nvme_devices(partitioned=True)
        if not nvme_devices:
            logger.info(f"{self.name} - Did not find any NVMe drive, skipping")
            self._done()
            return []
        nvme_dev = nvme_devices[0]
        partition = self._get_device_partition(nvme_dev)
        if not partition:
            err_msg = f"{self.name} - Expected {nvme_dev} to have been partitioned but it is not!"
            logger.error(err_msg)
            raise Exception(err_msg)
        mount_point = self._create_mount_point(partition)
        self._mount_partition(partition, mount_point)
        self._reorganize_directories(mount_point)

        self._done()
        logger.info(f"{self.name} - Finished execution")
        return []

    def _create_mount_point(self, partition: str) -> pathlib.Path:
        name = pathlib.Path(partition).name
        root = pathlib.Path('/mnt')
        mount_point = root / name
        logger.debug(f"{self.name} - Creating mount point {str(mount_point)}")
        mount_point.mkdir()
        logger.info(f"{self.name} - Created mount point {str(mount_point)}")
        return mount_point

    def _mount_partition(self, partition: str, mount_point: pathlib.Path) -> None:
        logger.debug(f"{self.name} - Mounting {partition} to {str(mount_point)}")
        post_exception_action = lambda: SystemUtils.execute_command(["umount", "-f", str(mount_point)], timeout=500)
        SystemUtils.execute_command_with_retry(["mount", partition, str(mount_point)], post_exception_action=post_exception_action)
        logger.info(f"{self.name} - Mounted {partition} to {str(mount_point)}")

    def _reorganize_directories(self, mount_point: pathlib.Path) -> None:
        logger.debug(f"{self.name} - Moving /srv/hops/hopsdata to {mount_point}")
        hopsdata = pathlib.Path('/srv/hops/hopsdata')
        real_hopsdata = hopsdata.resolve()
        cmd = f"mv -f {str(real_hopsdata)} {str(mount_point)}"
        # We have to use mv because Python shutil does not preserve metadata
        # when moving across different filesystems
        SystemUtils.execute_command(cmd.split(" "), timeout=600)
        logger.info(f"{self.name} - Moved /srv/hops/hopsdata to {mount_point}")

        new_hopsdata = mount_point / 'hopsdata'
        real_hopsdata.symlink_to(new_hopsdata)
        logger.info(f"{self.name} - symlinked {str(real_hopsdata)} to {str(new_hopsdata)}")
        logger.info(f"{self.name} - Finished moving hopsdata to NVMe drive")

        logger.debug(f"{self.name} - Moving /srv/hops/docker to {mount_point}")
        docker_dir = pathlib.Path('/srv/hops/docker')
        real_docker_dir = docker_dir.resolve()
        docker_size = sum(f.stat().st_size for f in real_docker_dir.glob('**/*') if f.is_file())
        if docker_size <= 100 * 1024 * 1024:
            is_docker_running = SystemdUtils.is_active('docker')
            logger.debug(f"{self.name} - Is docker service running? {is_docker_running}")
            if is_docker_running:
                SystemdUtils.stop('docker')

            cmd = f"mv -f {str(real_docker_dir)} {str(mount_point)}"
            SystemUtils.execute_command(cmd.split(" "), timeout=1800)

            new_docker_dir = mount_point / 'docker'
            real_docker_dir.symlink_to(new_docker_dir)
            logger.info(f"{self.name} - symlinked {str(real_docker_dir)} to {str(new_docker_dir)}")
            if is_docker_running:
                logger.debug(f"{self.name} - Starting docker service")
                SystemdUtils.start('docker')
            logger.info(f"{self.name} - Finished moving docker to NVMe drive")
        else:
            logger.info(f"{self.name} - Skip moving docker directory to NVMe drive because it is too big")


    def rollback(self) -> None:
        pass
